// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (C) 2017 MediaTek Inc.
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/rtc.h>

#include <mt-plat/mtk_ccci_common.h> /* exec_ccci_kern_func_by_md_id */

#if defined(CONFIG_MTK_WATCHDOG) && defined(CONFIG_MTK_WD_KICKER)
#include <mt-plat/mtk_wd_api.h> /* ap wdt related definitons */
#endif

#include <trace/events/mtk_idle_event.h>

#include <mtk_idle.h>
#include <mtk_idle_internal.h>
#include <mtk_spm_suspend_internal.h>
#include <mtk_spm_resource_req.h>

#include "mtk_spm_internal.h"
#include "pwr_ctrl.h"


#define MTK_IDLE_GS_DUMP_READY	(1)

#if MTK_IDLE_GS_DUMP_READY
/* NOTE: Check golden setting dump header file for each project */
#include <mtk_power_gs_api.h>
#endif

/* FIXME: IT with vcorefs ? */
void __attribute__((weak)) dvfsrc_md_scenario_update(bool suspend) {}

/********************************************************************
 * dp/so3/so pcm_flags and pcm_flags1
 *******************************************************************/

static unsigned int idle_pcm_flags[NR_IDLE_TYPES] = {
	[IDLE_TYPE_DP] =
		/* SPM_FLAG_DISABLE_CPU_PDN | */
		SPM_FLAG_DISABLE_INFRA_PDN |
		/* SPM_FLAG_DISABLE_DDRPHY_PDN | */
		SPM_FLAG_DISABLE_VCORE_DVS |
		SPM_FLAG_DISABLE_VCORE_DFS |
		SPM_FLAG_KEEP_CSYSPWRUPACK_HIGH |
		#if !defined(CONFIG_MTK_TINYSYS_SSPM_SUPPORT)
		SPM_FLAG_DISABLE_SSPM_SRAM_SLEEP |
		#endif
		SPM_FLAG_DEEPIDLE_OPTION,

	[IDLE_TYPE_SO3] =
		SPM_FLAG_DISABLE_INFRA_PDN |
		SPM_FLAG_DISABLE_VCORE_DVS |
		SPM_FLAG_DISABLE_VCORE_DFS |
		SPM_FLAG_KEEP_CSYSPWRUPACK_HIGH |
		#if !defined(CONFIG_MTK_TINYSYS_SSPM_SUPPORT)
		SPM_FLAG_DISABLE_SSPM_SRAM_SLEEP |
		#endif
		SPM_FLAG_SODI_OPTION |
		SPM_FLAG_ENABLE_SODI3,

	[IDLE_TYPE_SO] =
		SPM_FLAG_DISABLE_INFRA_PDN |
		SPM_FLAG_DISABLE_VCORE_DVS |
		SPM_FLAG_DISABLE_VCORE_DFS |
		SPM_FLAG_KEEP_CSYSPWRUPACK_HIGH |
		#if !defined(CONFIG_MTK_TINYSYS_SSPM_SUPPORT)
		SPM_FLAG_DISABLE_SSPM_SRAM_SLEEP |
		#endif
		SPM_FLAG_SODI_OPTION,

	[IDLE_TYPE_RG] = 0,
};

static unsigned int idle_pcm_flags1[NR_IDLE_TYPES] = {
	[IDLE_TYPE_DP] =
		SPM_FLAG1_ENABLE_BIG_BUCK_OFF |
		SPM_FLAG1_ENABLE_BIG_BUCK_ON |
		SPM_FLAG1_FORCE_CPU_BUCK_OFF |
		SPM_FLAG1_DISABLE_VS1_VOTER |
		SPM_FLAG1_DISABLE_VS2_VOTER,

	[IDLE_TYPE_SO3] =
		SPM_FLAG1_ENABLE_BIG_BUCK_OFF |
		SPM_FLAG1_ENABLE_BIG_BUCK_ON |
		SPM_FLAG1_FORCE_CPU_BUCK_OFF |
		SPM_FLAG1_DISABLE_VS1_VOTER |
		SPM_FLAG1_DISABLE_VS2_VOTER,

	[IDLE_TYPE_SO] =
		SPM_FLAG1_ENABLE_BIG_BUCK_OFF |
		SPM_FLAG1_ENABLE_BIG_BUCK_ON |
		SPM_FLAG1_FORCE_CPU_BUCK_OFF |
		SPM_FLAG1_DISABLE_VS1_VOTER |
		SPM_FLAG1_DISABLE_VS2_VOTER,

	[IDLE_TYPE_RG] = 0,
};


/********************************************************************
 * dp/so3/so pwrctrl variables
 *******************************************************************/

struct pwr_ctrl pwrctrl_dp;
struct pwr_ctrl pwrctrl_so3;
struct pwr_ctrl pwrctrl_so;

static struct pwr_ctrl *get_pwrctrl(int idle_type)
{
	return idle_type == IDLE_TYPE_DP ? &pwrctrl_dp :
		idle_type == IDLE_TYPE_SO3 ? &pwrctrl_so3 :
		idle_type == IDLE_TYPE_SO ? &pwrctrl_so : NULL;
}

static void mtk_idle_gs_dump(int idle_type)
{
	#if MTK_IDLE_GS_DUMP_READY
	if (idle_type == IDLE_TYPE_DP)
		mt_power_gs_dump_dpidle(GS_ALL);
	else if (idle_type == IDLE_TYPE_SO3 || idle_type == IDLE_TYPE_SO)
		mt_power_gs_dump_sodi3(GS_ALL);
	#endif
}

/********************************************************************
 * dp/so3/so trigger wfi
 *******************************************************************/

static void print_ftrace_tag(int idle_type, int cpu, int enter)
{
#if MTK_IDLE_TRACE_TAG_ENABLE
	switch (idle_type) {
	case IDLE_TYPE_DP:
		trace_dpidle_rcuidle(cpu, enter);
		break;
	case IDLE_TYPE_SO:
		trace_sodi_rcuidle(cpu, enter);
		break;
	case IDLE_TYPE_SO3:
		trace_sodi3_rcuidle(cpu, enter);
		break;
	default:
		break;
	}
#endif
}

int mtk_idle_trigger_wfi(int idle_type, unsigned int idle_flag, int cpu)
{
	int spm_dormant_sta = 0;
	struct pwr_ctrl *pwrctrl;
	unsigned int cpuidle_mode[NR_IDLE_TYPES] = {
		[IDLE_TYPE_DP] = MTK_DPIDLE_MODE,
		[IDLE_TYPE_SO3] = MTK_SODI3_MODE,
		[IDLE_TYPE_SO] = MTK_SODI_MODE,
	};
	unsigned int enter_flag[NR_IDLE_TYPES] = {
		[IDLE_TYPE_DP] = SPM_ARGS_DPIDLE,
		[IDLE_TYPE_SO3] = SPM_ARGS_SODI,
		[IDLE_TYPE_SO] = SPM_ARGS_SODI,
	};
	unsigned int leave_flag[NR_IDLE_TYPES] = {
		[IDLE_TYPE_DP] = SPM_ARGS_DPIDLE_FINISH,
		[IDLE_TYPE_SO3] = SPM_ARGS_SODI_FINISH,
		[IDLE_TYPE_SO] = SPM_ARGS_SODI_FINISH,
	};

	/* Dump low power golden setting */
	if (idle_flag & MTK_IDLE_LOG_DUMP_LP_GS)
		mtk_idle_gs_dump(idle_type);

	pwrctrl = get_pwrctrl(idle_type);

	if (!pwrctrl) {
		pr_info("Error: pwrctrl is NULL!\n");
		return 0;
	}


	print_ftrace_tag(idle_type, cpu, 1);

	if (is_cpu_pdn(pwrctrl->pcm_flags))
		spm_dormant_sta = mtk_enter_idle_state(cpuidle_mode[idle_type]);
	else {
		mt_secure_call(MTK_SIP_KERNEL_SPM_ARGS
					, enter_flag[idle_type], 0, 0, 0);
		mt_secure_call(MTK_SIP_KERNEL_SPM_LEGACY_SLEEP
					, 0, 0, 0, 0);
		mt_secure_call(MTK_SIP_KERNEL_SPM_ARGS
					, leave_flag[idle_type], 0, 0, 0);
	}

	print_ftrace_tag(idle_type, cpu, 0);

	if (spm_dormant_sta < 0)
		pr_info("mtk_enter_idle_state(%d) ret %d\n",
			cpuidle_mode[idle_type], spm_dormant_sta);

	return spm_dormant_sta;
}


/********************************************************************
 * dp/so3/so setup/cleanup pcm
 *******************************************************************/

static int smc_id[NR_IDLE_TYPES] = {
	[IDLE_TYPE_DP] = MTK_SIP_KERNEL_SPM_DPIDLE_ARGS,
	[IDLE_TYPE_SO3] = MTK_SIP_KERNEL_SPM_SODI_ARGS,
	[IDLE_TYPE_SO] = MTK_SIP_KERNEL_SPM_SODI_ARGS,
};

static void spm_idle_pcm_setup_before_wfi(
	int idle_type, struct pwr_ctrl *pwrctrl, unsigned int op_cond)
{
	unsigned int resource_usage = 0;

	resource_usage = (op_cond & MTK_IDLE_OPT_SLEEP_DPIDLE) ?
		spm_get_resource_usage_by_user(SPM_RESOURCE_USER_SCP)
		: spm_get_resource_usage();

	mt_secure_call(smc_id[idle_type], pwrctrl->pcm_flags,
		pwrctrl->pcm_flags1, resource_usage, 0);

	/* [sodi3]
	 * ap wdt works fine without f26m and no need to enable pcm wdt
	 *
	 */
	if (idle_type == IDLE_TYPE_SO3)
		mt_secure_call(MTK_SIP_KERNEL_SPM_PWR_CTRL_ARGS
			, SPM_PWR_CTRL_SODI3, PW_WDT_DISABLE, 1, 0);

	/* [sleep dpidle] bypass wake_src and pcm timer value */
	if (op_cond & MTK_IDLE_OPT_SLEEP_DPIDLE)
		mt_secure_call(MTK_SIP_KERNEL_SPM_SLEEP_DPIDLE_ARGS
			, pwrctrl->timer_val, pwrctrl->wake_src, 0, 0);
}

static void spm_idle_pcm_setup_after_wfi(
	int idle_type, struct pwr_ctrl *pwrctrl, unsigned int op_cond)
{

}


/********************************************************************
 * dp/so3/so main flow by chip
 *******************************************************************/

static unsigned int slp_dp_timer_val;
static unsigned int slp_dp_wake_src;
static unsigned int slp_dp_last_wr = WR_NONE;
static unsigned long flags;
#if defined(CONFIG_MTK_WATCHDOG) && defined(CONFIG_MTK_WD_KICKER)
static struct wd_api *wd_api;
static int wd_ret;
#endif

bool mtk_idle_resource_pre_process(void)
{
	if (mtk_dpidle_enabled() &&
		mtk_spm_arch_type_get() == MTK_SPM_ARCH_RESOURCE_ORIENTED)
		return (spm_get_resource_usage() & SPM_RESOURCE_CK_26M);
	else
		return 0;
}

void mtk_idle_pre_process_by_chip(
	int idle_type, int cpu, unsigned int op_cond, unsigned int idle_flag)
{
	struct pwr_ctrl *pwrctrl = get_pwrctrl(idle_type);
	unsigned int pcm_flags;
	unsigned int pcm_flags1;

	if (!pwrctrl) {
		pr_info("Error: pwrctrl is NULL!\n");
		return;
	}

	/* get pcm_flags and update if needed */
	pcm_flags = idle_pcm_flags[idle_type];
	pcm_flags1 = idle_pcm_flags1[idle_type];

	/* [sleep dpidle only] */
	if (op_cond & MTK_IDLE_OPT_SLEEP_DPIDLE) {

		/* backup timer_val and wake_src */
		slp_dp_timer_val = pwrctrl->timer_val;
		slp_dp_wake_src = pwrctrl->wake_src;

		/* setup sleep wake_src and timer_val */
		pwrctrl->timer_val =
			32768 * __spm_get_wake_period(-1, slp_dp_last_wr);

		/* Get sleep wake_src for sleep dpidle */
		pwrctrl->wake_src = spm_get_sleep_wakesrc();

		/* pre watch dog config */
#if defined(CONFIG_MTK_WATCHDOG) && defined(CONFIG_MTK_WD_KICKER)

		wd_ret = get_wd_api(&wd_api);
		if (!wd_ret && wd_api) {
			wd_api->wd_spmwdt_mode_config(WD_REQ_EN
							, WD_REQ_RST_MODE);
			wd_api->wd_suspend_notify();
		} else {
			aee_sram_printk("FAILED TO GET WD API\n");
			pr_info("[IDLE] FAILED TO GET WD API\n");
		}
#else
		SMC_CALL(ARGS, SPM_ARGS_PCM_WDT, 1, 30);
#endif
	}

	/* initialize pcm_flags/pcm_flags1 */
	__spm_set_pwrctrl_pcm_flags(pwrctrl, pcm_flags);
	__spm_set_pwrctrl_pcm_flags1(pwrctrl, pcm_flags1);

	/* lock spm spin_lock */
	spin_lock_irqsave(&__spm_lock, flags);

	/* mask irq and backup cirq */
	mtk_spm_irq_backup();

	__spm_sync_pcm_flags(pwrctrl);

	/* setup vcore dvfs before idle scenario */
	dvfsrc_md_scenario_update(1);

	/* setup spm */
	spm_idle_pcm_setup_before_wfi(idle_type, pwrctrl, op_cond);
}

static unsigned int mtk_dpidle_output_log(
	int idle_type, const struct wake_status *wakesta,
	unsigned int op_cond, unsigned int idle_flag);

static unsigned int mtk_sodi_output_log(
	int idle_type, const struct wake_status *wakesta,
	unsigned int op_cond, unsigned int idle_flag);

typedef unsigned int (*mtk_idle_log_t) (
	int idle_type, const struct wake_status *wakesta,
	unsigned int op_cond, unsigned int idle_flag);

static mtk_idle_log_t mtk_idle_log[NR_IDLE_TYPES] = {
	[IDLE_TYPE_DP] = mtk_dpidle_output_log,
	[IDLE_TYPE_SO3] = mtk_sodi_output_log,
	[IDLE_TYPE_SO] = mtk_sodi_output_log,
};

void mtk_idle_post_process_by_chip(
	int idle_type, int cpu, unsigned int op_cond, unsigned int idle_flag)
{
	struct pwr_ctrl *pwrctrl = get_pwrctrl(idle_type);
	struct wake_status wakesta;
	unsigned int wr = WR_NONE;

	if (!pwrctrl) {
		pr_info("Error: pwrctrl is NULL!\n");
		return;
	}

	if (unlikely(idle_type < 0 || idle_type >= NR_IDLE_TYPES))
		return;

	/* get spm info */
	__spm_get_wakeup_status(&wakesta);

	/* save ap and 26M's off counter and duration */
	__spm_save_ap_sleep_info(&wakesta);
	if (idle_type == IDLE_TYPE_SO3)
		__spm_save_26m_sleep_info();

	/* clean up spm */
	spm_idle_pcm_setup_after_wfi(idle_type, pwrctrl, op_cond);

	/* setup vcore dvfs after idle scenario */
	dvfsrc_md_scenario_update(0);

	/* print log */
	wr = mtk_idle_log[idle_type](idle_type, &wakesta, op_cond, idle_flag);

	/* unmask irq and restore cirq */
	mtk_spm_irq_restore();

	/* unlock spm spin_lock */
	spin_unlock_irqrestore(&__spm_lock, flags);

	/* [sleep dpidle only] */
	if (op_cond & MTK_IDLE_OPT_SLEEP_DPIDLE) {
		/* post watch dog config */
#if defined(CONFIG_MTK_WATCHDOG) && defined(CONFIG_MTK_WD_KICKER)
		if (!wd_ret) {
			if (!pwrctrl->wdt_disable)
				wd_api->wd_resume_notify();
			else {
				aee_sram_printk(
					"pwrctrl->wdt_disable %d\n",
						pwrctrl->wdt_disable);
				pr_info(
					"[SPM] pwrctrl->wdt_disable %d\n",
						pwrctrl->wdt_disable);
			}

			wd_api->wd_spmwdt_mode_config(WD_REQ_DIS
							, WD_REQ_RST_MODE);
		}
#else
		SMC_CALL(ARGS, SPM_ARGS_PCM_WDT, 0, 0);
#endif

		/* restore timer_val and wake_src */
		pwrctrl->timer_val = slp_dp_timer_val;
		pwrctrl->wake_src = slp_dp_wake_src;

		/* backup wakeup reason */
		slp_dp_last_wr = wr;
	}

	/* patch: rekick vcorefs */
	if (wr == WR_ABORT)
		rekick_vcorefs_scenario();
}

unsigned int get_slp_dp_last_wr(void)
{
	return slp_dp_last_wr;
}
/********************************************************************
 * mtk idle output log
 *******************************************************************/
#define IDLE_TIMER_OUT_CRITERIA (32)    /* 1 ms (32k/sec)*/
#define IDLE_PRINT_LOG_DURATION (5000)  /* 5 seconds */

static bool check_print_log_duration(void)
{
	static unsigned long  pre_time;
	unsigned long  cur_time;
	bool ret = false;
	struct timespec ts;
	struct rtc_time tm;

	cur_time = spm_get_current_time_ms();
	ret = (cur_time - pre_time > IDLE_PRINT_LOG_DURATION);

	if (ret) {
		pre_time = cur_time;

		getnstimeofday(&ts);
		rtc_time_to_tm(ts.tv_sec, &tm);
		pr_info(
		"[name:spm&][SPM] idle: %d-%02d-%02d %02d:%02d:%02d.%09lu UTC\n",
			tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
			tm.tm_hour, tm.tm_min, tm.tm_sec, ts.tv_nsec);
	}

	return ret;
}

/* Print output log - Deep Idle ------------------------ */

static unsigned int mtk_dpidle_output_log(
	int idle_type, const struct wake_status *wakesta,
	unsigned int op_cond, unsigned int idle_flag)
{
	bool print_log = false;
	unsigned int wr = WR_NONE;

	/* No log for latency profiling case */
	if (idle_flag & MTK_IDLE_LOG_DISABLE)
		return WR_NONE;

	/* [sleep dpidle] directly print */
	if (op_cond & MTK_IDLE_OPT_SLEEP_DPIDLE)
		return __spm_output_wake_reason(wakesta, true, "sleep_dpidle");

	if (!(idle_flag & MTK_IDLE_LOG_REDUCE)) {
		print_log = true;
	} else {
		if (wakesta->is_abort != 0 || wakesta->r12 == 0) {
			print_log = true;
		} else if (check_print_log_duration()) {
			print_log = true;
		} else if (wakesta->timer_out <= IDLE_TIMER_OUT_CRITERIA) {
			print_log = true;
			if (wakesta->r12 == R12_AP2AP_PEER_WAKEUP_EVENT)
				print_log = false;
		}
	}

	if (print_log) {
		printk_deferred("[name:spm&]Power/swap op_cond = 0x%x\n",
				op_cond);
		wr = __spm_output_wake_reason(
			wakesta, false, mtk_idle_name(idle_type));
		if (idle_flag & MTK_IDLE_LOG_RESOURCE_USAGE)
			spm_resource_req_dump();
	}

	return wr;
}

/* Print output log - SODI3 / SODI --------------------- */

static unsigned int mtk_sodi_output_log(
	int idle_type, const struct wake_status *wakesta,
	unsigned int op_cond, unsigned int idle_flag)
{
	bool print_log = false;
	unsigned int wr = WR_NONE;

	/* No log for latency profiling case */
	if (idle_flag & MTK_IDLE_LOG_DISABLE)
		return WR_NONE;

	if (!(idle_flag & MTK_IDLE_LOG_REDUCE)) {
		print_log = true;
	} else {
		if (wakesta->is_abort != 0 || wakesta->r12 == 0) {
			print_log = true;
		} else if (check_print_log_duration()) {
			print_log = true;
		} else if (wakesta->timer_out <= IDLE_TIMER_OUT_CRITERIA) {
			print_log = true;
			if (wakesta->r12 == R12_AP2AP_PEER_WAKEUP_EVENT)
				print_log = false;
		}
	}

	if (print_log) {
		printk_deferred("[name:spm&]Power/swap op_cond = 0x%x\n",
				op_cond);
		wr = __spm_output_wake_reason(
			wakesta, false, mtk_idle_name(idle_type));
		if (idle_flag & MTK_IDLE_LOG_RESOURCE_USAGE)
			spm_resource_req_dump();
	}

	return wr;
}



